# # 10-fold validation 

import sys, getopt, collections

# # Viterbi ## use global prob matrix tagtoword
def Viterbi(observation,tagtransgraph,starttaggraph,mintagtoword):
  globaltags = {'ADV':1,'NOUN':2,'ADP':3,'.':4,'PRON':5,'VERB':6,'X':7,'CONJ':8,'ADJ':9}
  lenglobaltags = len(globaltags)
  lenobservation = len(observation)
  lentagtransgraph = len(tagtransgraph)
  viterbimatrix = [[None for _ in range(lenobservation)] for _ in range(lenglobaltags+2)]
  backpointer = [[None for _ in range(lenobservation)] for _ in range(lenglobaltags+2)]
  for i,state in enumerate(globaltags):
    viterbimatrix[globaltags[state]][0] = starttaggraph[('START',state)]*(tagtoword[(state,observation[0])]+0.00000001)
    backpointer[globaltags[state]][0] = 0
  for i in range(lenobservation-1):
    for j,state in enumerate(globaltags):
      #print viterbimatrix
      viterbimatrix[globaltags[state]][i+1] = max([viterbimatrix[globaltags[s]][i]*tagtransgraph[(s,state)]*(tagtoword[(state,observation[i+1])]+0.00000001) for s in globaltags])
      argmaxS = 'NONE'
      maxS = -1
      for s in globaltags:
        newmax = viterbimatrix[globaltags[s]][i]*tagtransgraph[(s,state)]
        if newmax>maxS:
          maxS = newmax
          argmaxS = s
      backpointer[globaltags[state]][i+1] = argmaxS 
  #viterbimatrix[lenglobaltags+1][lenobservation] = max([viterbimatrix[globaltags[s]][lenobservation]* for s in globaltags])
  argmaxGlobal = 'NONE'
  maxGlobal = -1
  for i in globaltags:
    if viterbimatrix[globaltags[i]][lenobservation-1] > maxGlobal:
      maxGlobal = viterbimatrix[globaltags[i]][lenobservation-1]
      argmaxGlobal = i

  # # get tag path
  tagpath = []
  tagpath.append(argmaxGlobal)
  for i in range(lenobservation-1):
    #print argmaxGlobal
    tagpath.append(backpointer[globaltags[argmaxGlobal]][lenobservation-i-1])
    argmaxGlobal = backpointer[globaltags[argmaxGlobal]][lenobservation-i-1]
  return tagpath[::-1]


def danishsyntaximprove(taglist,wordlist):
  improved = []
  return improved

# # improve definitive {. : , ?} etc 
def generalsyntaximprove(taglist,wordlist):
  improved = []
  for i,item in enumerate(wordlist):
    if((item.strip() in ['.',',',':',';','"',"'",'?','-','(',')','!','#','$','%','^','&','*','+','=','/','{','}','[',']','@']) and (taglist[i] != ".")):
      improved.append('.')
      print "improved ",taglist[i]," tag: ",taglist[i]
    elif((item.strip() not in ['.',',',':',';','"',"'",'?','-','(',')','!','#','$','%','^','&','*','+','=','/','{','}','[',']','@']) and taglist[i]=="."):
      improved.append('ADJ')
      print "improved reverse: ",item," tag: ",taglist[i]," word: ",wordlist[i]
    else:
      improved.append(taglist[i])
  return improved


###################################################################

train_file=""
test_file=""
readable = False
try:
  opts, args = getopt.getopt(sys.argv[1:], "h", ["help", "readable", "train=", "test="])
except getopt.GetoptError:
  print "Usage: example.py [--readable]"
  sys.exit(2)
for opt, arg in opts:
  if opt in ("-h", "--help"):
    print "Usage: example.py [--readable]"
    sys.exit()
  elif opt in ("--readable"):                
    readable = True                 
  elif opt in ("--train"):                
    train_file = arg 
  elif opt in ("--test"):                
    test_file = arg 


###################################################################

# # 10-fold validation
def splitsentence(filename):
  sentencearray = []
  for line in open(str(filename)):
    sentencearray.append(line)
  return sentencearray

def traintestprocess(fullarray,position1,position2):
  lenarray = len(fullarray)
  trainlist = fullarray[0:position1*lenarray/10] + fullarray[position2*lenarray/10:]
  testlist = fullarray[position1*lenarray/10:position2*lenarray/10]
  return [trainlist,testlist]

def testlinefactory(testarray):
  testwordarray = []
  testtagarray = []
  for line in testarray:
    taglist = [x.rsplit('/',1) for x in line.split()]
    tempword = []
    temptag = []
    for tag in taglist:
      tempword.append(tag[0])
      temptag.append(tag[1])
    testwordarray.append(tempword)
    testtagarray.append(temptag)
  return [testwordarray,testtagarray]

###################################################################

## splitted arrays
full_lines = []
train_lines = []
test_lines = []

# # Transition of tags by pairs
transtotal = 0
linecount = 0
transtag = collections.defaultdict(int)
starttag = collections.defaultdict(int)
starttagprob = collections.defaultdict(float)
transtagprob = collections.defaultdict(float)

full_lines = splitsentence(train_file)
templist = traintestprocess(full_lines,0,1)
train_lines = templist[0]
test_lines = templist[1]


for line in train_lines:
  taglist = [x.rsplit('/',1) for x in line.split()]
  starttag[('START',taglist[0][1])]+=1
  linecount+=1
  taglen = len(taglist)
  for index,tagpair in enumerate(taglist):
    if index<taglen-1:
      nexttag = taglist[index+1]
      transtag[(tagpair[1],nexttag[1])]+=1
      transtotal+=1

# # Add-one smoothing
typelendivisor = len(transtag)+transtotal
starttagdivisor = len(starttag)+linecount

print len(train_lines)
print len(full_lines)
print len(test_lines)

# # get probability
for item in transtag:
  transtagprob[item]=(transtag[item]+1.0)/typelendivisor
for item in starttag:
  starttagprob[item]=(starttag[item]+1.0)/starttagdivisor


# # tag to words
countword = 0
tagtoword = collections.defaultdict(float)
tagworddict = collections.defaultdict(int)
for line in train_lines:
  for token,tag in [x.rsplit('/',1) for x in line.split()]:
    tagtoword[(tag,token)]+=1
    tagworddict[tag]+=1
    countword+=1
for item in tagtoword:
  tagtoword[item]=tagtoword[item]*1.0/tagworddict[item[0]]

temptestarray = testlinefactory(test_lines)
testwordlist = temptestarray[0]
testtaglist = temptestarray[1]

mintagtoword = min(tagtoword.values())

###################################################################

validation_total = 0
validation_hit = 0
for index,line in enumerate(testwordlist):
  path = Viterbi(line,transtagprob,starttagprob,mintagtoword)
  path = generalsyntaximprove(path,line)
  validation_total+=len(testtaglist[index])
  for i,item in enumerate(path):
    if(item.strip() == testtaglist[index][i].strip()):
      validation_hit+=1
    else:
      #print item," :is not: ",testtaglist[index][i], " word: ", line[i]
      pass

print validation_total
print validation_hit
print validation_hit*1.0/validation_total








